/*
  +----------------------------------------------------------------------+
  | Copyright (c) The PHP Group                                          |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.01 of the PHP license,      |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_01.txt                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Authors: Rafal Somla <rafal.somla@oracle.com>                        |
  +----------------------------------------------------------------------+
*/

extern "C" {
#include "php_mysqlnd_krb.h"
}

#include "gssapi.h"

#include <array>
#include <sstream>

#pragma comment(lib, "gssapi64.lib")
#pragma comment(lib, "krb5_64.lib")

using namespace std;


/*
  Functions to interact with the krb5 layer to get a TGT from a Kerberos server.
*/

namespace krb {
  
  // Return a krb5 context, initializing it first if not yet done.

  const krb5_context& ctx()
  {
    static
    struct Lib_ctx
    {
      krb5_context m_ctx;

      Lib_ctx()
      {
        DBG_FMT("Krb5 context initialization");
        if (!krb5_init_context(&m_ctx))
          return;
        throw std::runtime_error{
          "Could not initialize Kerberos library contex"
        };
      }

      ~Lib_ctx()
      {
        DBG_FMT("Krb5 context deinitialization");
        krb5_free_context(m_ctx);
      }
    }
    instance;

    return instance.m_ctx;
  }


  const char* get_error_msg(uint32_t err)
  {
    static std::array<char, 512> buf;
    const char *msg = krb5_get_error_message(ctx(), err);
    strncpy(buf.data(), msg, buf.size());
    krb5_free_error_message(ctx(), msg);
    return buf.data();
  }

  void check_err(krb5_error_code ret) {
    if (ret)
      throw get_error_msg(ret);
  }

  void acquire_credentials(string upn, string password)
  {
    // See: https://web.mit.edu/kerberos/krb5-latest/doc/appdev/init_creds.html

    DBG_FMT("Acquiring Kerberos credentials (TGT): upn='%s', password='%s'"
      , upn.c_str()
      , password.c_str()
    );    

    krb5_creds credentials;
    krb5_principal ukrb;
    krb5_error_code ret;
    krb5_ccache cache = {0};

    try {

      check_err(krb5_parse_name(ctx(), upn.c_str(), &ukrb));

      // Get the default cahce for the UPN

      // TODO: Need to call krb5_cc_close on returned cache?

      ret = krb5_cc_cache_match(ctx(), ukrb, &cache);

      /*
        Note: According to GSSAPI specs, KRB5_CC_NOTFOUND should be returned 
        if cache is not found. But on Win some random error is returned 
        instead. So we treat all errors as indication that the cache could 
        not be found.
      */

      if (ret) //KRB5_CC_NOTFOUND == ret)
      {
        DBG_FMT("No cache found for UPN, initializing the default cache");

        // TODO? Use tmp. MEMORY cache instead?
        // see: https://web.mit.edu/kerberos/krb5-latest/doc/appdev/refs/api/krb5_cc_new_unique.html

        // see: https://web.mit.edu/kerberos/krb5-latest/doc/appdev/refs/api/krb5_cc_default_name.html#c.krb5_cc_default_name

        check_err(krb5_cc_default(ctx(), &cache));

        // set default principal for the cache
        // TODO: Instead of blindly overwriting the default UPN, check that it is correct
        check_err(krb5_cc_initialize(ctx(), cache, ukrb));
      }
      //else
      //  check_err(ret);

      DBG_FMT("Using cache: %s:%s"
        , krb5_cc_get_type(ctx(), cache)
        , krb5_cc_get_name(ctx(), cache)
      );

      // Options for krb5_get_init_creds_password()

      struct Options
      {
        krb5_get_init_creds_opt *ptr = nullptr;

        Options()
        {
          /*
            Note: we ignore errors assuming that _opt_free() needs to be 
            called if this function  sets the pointer.
          */
          (void)krb5_get_init_creds_opt_alloc(ctx(), &ptr);
        }

        ~Options()
        {
          if (!ptr)
            return;
          // Note: ignore errors here.
          (void)krb5_get_init_creds_opt_free(ctx(), ptr);
        }

        void set(krb5_ccache cache)
        {
          /*
            These options set the input cache which is used to read initial 
            parameters and the output cache where the obtained credentials will 
            be stored.

            Note: ignore error if input cache could not be set because all 
            should work fine also in this case.
          */

          (void)krb5_get_init_creds_opt_set_in_ccache(ctx(), ptr, cache);
          check_err(krb5_get_init_creds_opt_set_out_ccache(ctx(), ptr, cache));
        }
      }
      opts;

      opts.set(cache);

      check_err(krb5_get_init_creds_password(ctx()
        , &credentials
        , ukrb
        , password.c_str()
        , nullptr, nullptr // prompter function and its data 
        , 0 // start time
        , nullptr // send request to ticket-granting service
        , opts.ptr
      ));

    }
    catch(const char *err)
    {
      throw std::runtime_error{std::string{"Failed to obtain TGT: "} + err};
    }

    DBG_FMT("Credentials acquired");
  }

} // krb namespace



namespace {

  // The GSSAPI context used to perform authentication
  optional<gssapi::Context> ctx;


/* {{{ mysqlnd_ldap_sasl_get_auth_data 

  Compute response to server's authentication challenge.

  The challenge data is passed as `auth_plugin_data` and a pointer to malloc() 
  allocated response should be returned, setting `auth_data_len` to the length 
  of the response.

  For Kerberos the initial challenge contains server's SPN and the Kerberos 
  domain of the user account. The response is the first token generated 
  by GSS/Kerberos handshake for the user principal and SPN target. The required 
  Kerberos tickets should be in the chache or, if it is not the case, an 
  attempt to acquire them using provided password is made.

  FIXME: We assume here that the handshake will have only one iteration (in 
  testing this seems to be the case). But GSSAPI allows multiple iterations and 
  to be on the safe side we should be ready to handle that. However it is not 
  clear if mysqlnd authentication logic supports multiple handshake rounds.

  FIXME: Think if this implementation is thread safe - does mysqlnd guarantee 
  that two handshakes will never happen concurently?

*/

static zend_uchar *
mysqlnd_kerberos__get_auth_data(
        struct st_mysqlnd_authentication_plugin * self,
        size_t * auth_data_len,
        MYSQLND_CONN_DATA * conn, const char * const user, const char * const passwd,
        const size_t passwd_len, zend_uchar * auth_plugin_data, const size_t auth_plugin_data_len,
        const MYSQLND_SESSION_OPTIONS * const session_options,
        const MYSQLND_PFC_DATA * const pfc_data,
        const zend_ulong mysql_flags
)
try {
    DBG_ENTER("mysqlnd_kerberos_get_auth_data");
    DBG_FMT(PHP_MYSQLND_KRB_VERSION);

    using gssapi::Buffer;

    if(ctx)
    {
      throw std::runtime_error{
        "Not implemented: second handshake round!"
      };
      DBG_RETURN(nullptr);
    }

    // The handshake did not start yet, we need to create context.

    DBG_FMT("  user: %s", user ? user : "<empty>");
    DBG_FMT("passwd: %s", passwd ? passwd : "<empty>");

    /*
      auth_plugin_data format:
        SPN length (2 bytes) + SPN string
        realm length (2 bytes) + realm string

      Example:

      2C 00 6D 79 73 71 6C 5F  73 65 72 76 69 63 65 2F  |  ,.mysql_service/ 
      6B 65 72 62 65 72 6F 73  5F 61 75 74 68 5F 68 6F  |  kerberos_auth_ho 
      73 74 40 4D 59 53 51 4C  2E 4C 4F 43 41 4C 0B 00  |  st@MYSQL.LOCAL.. 
      4D 59 53 51 4C 2E 4C 4F  43 41 4C                 |  MYSQL.LOCAL 
    */
    //printf("auth_plugin_data:\n");
    //hex_dump(auth_plugin_data, auth_plugin_data_len);

    zend_uchar *pos = auth_plugin_data;
    size_t len;

    len = (size_t)*pos++;
    len += ((size_t)*pos++) << 8;

    if (len + 2 > auth_plugin_data_len)
      throw std::invalid_argument{"String length overflow in server challenge"};

    std::string spn = {pos, pos+len};

    pos += len;

    size_t len1;
    len1 = (size_t)*pos++;
    len1 += ((size_t)*pos++) << 8;

    if (len + len1 + 4 > auth_plugin_data_len)
      throw std::invalid_argument{"String length overflow in server challenge"};

    std::string realm = {pos, pos+len1};

    DBG_FMT("  SPN: %s", spn.c_str());
    DBG_FMT("realm: %s", realm.c_str());

    std::string upn = std::string{user} + "@" + realm;

    DBG_FMT("  UPN: %s", upn.c_str());

    try {
      /*
        Create new context.

        Function passed as the 3rd argument is used to fetch TGT from Kerberos 
        server in case it can not be found in the cache.
      */

      ctx.emplace(spn, upn, [passwd](std::string upn) {
        krb::acquire_credentials(upn, passwd);
      });

      // DBG_FMT("context created, generating token");

      // The initial challenge
      auto challenge = ctx->gen_token();

      if (!challenge)
      {
        DBG_FMT("No token generated.");
        DBG_RETURN(nullptr);
      }

      if (auth_data_len)
        *auth_data_len = challenge->length;

      /*
        Note: We need to copy the buffer allocated by GSSAPI library because 
        the returned memory will be freed by mysqlnd using standard free() 
        call while the GSSAPI buffer must be freed using GSSAPI calls.
      */

      auto *ptr = (zend_uchar*)malloc(challenge->length);
      memcpy(ptr, challenge->data(), challenge->length);

      DBG_RETURN(ptr);
    }
    catch (const std::exception &ex)
    {
      std::ostringstream err_msg;
      err_msg
      << "Kerberos authentication failed ("
      << "for UPN '" << upn << "'"
      << " and SPN '" << spn << "'"
      << "): " << ex.what();

      throw std::runtime_error{err_msg.str()};
    }
}
catch(const std::exception &ex)
{
    php_error(E_ERROR, ex.what());
    DBG_RETURN(NULL);
}
catch(...)
{
    php_error(E_ERROR, "Unknown error!");
    DBG_RETURN(NULL);
}


func_status
mysqlnd_kerberos__handle_server_response(
    st_mysqlnd_authentication_plugin * self,
		MYSQLND_CONN_DATA * conn,
		const zend_uchar * auth_plugin_data, size_t auth_plugin_data_len,
		const char * const passwd,
		const size_t passwd_len,
		char **new_auth_protocol, size_t *new_auth_protocol_len,
		zend_uchar **new_auth_protocol_data, size_t *new_auth_protocol_data_len
		)
{
  DBG_ENTER("mysqlnd_kerberos_server_response");

  // FIXME: Detect if more iterations are needed and destroy the 
  // context only if this is not the case.
  DBG_FMT("destroying gssapi context");
  ctx.reset();

  DBG_RETURN(PASS);
}


/*
  Plugin definition.
*/

struct st_mysqlnd_authentication_plugin
plugin =
{
  {
      MYSQLND_PLUGIN_API_VERSION,
      "auth_plugin_authentication_kerberos_client",
      MYSQLND_VERSION_ID, // TODO: verify if this is correct.
      PHP_MYSQLND_VERSION,
      "PHP License 3.01",
      "Rafal Somla <rafal.somla@oracle.com>",
      {
          NULL, /* no statistics , will be filled later if there are some */
          NULL, /* no statistics */
      },
      {
          NULL /* plugin shutdown */
      }
  },
  {/* methods */
      mysqlnd_kerberos__get_auth_data,
      mysqlnd_kerberos__handle_server_response
  }
};

} // anonymous namespace



/*
  Return pointer to the structure describing authentication plugin.

  Caveat: This plugin is registered with mysqlnd and it is deactivated 
  by mysqlnd when the latter is shut down. It can happen that this plugin DLL 
  is unloaded before mysqlnd shuts down. For that reason we can not return 
  a pointer to a static memory inside the DLL. Instead, we return pointer 
  to the structure dynamically allocated on the heap so that it will survive 
  unloading of this plugin DLL. 
  
  The down-side is that this allocated memory will never be freed thus leading 
  to a memory leak. However, this should not be a big problem because normally 
  the authentication plugin should be loaded/registered only once during PHP 
  life-time and the PHP process shuld terminate shortly after the moment when 
  the plugin is no longer needed (after shutting down mysqlnd).
*/

extern "C"
struct st_mysqlnd_plugin_header *mysqlnd_kerberos_auth_plugin()
{
  // Note: use static instance to allocate memory only once.

  static struct Once
  {
    st_mysqlnd_authentication_plugin *ptr
    = new st_mysqlnd_authentication_plugin{plugin};
  }
  instance;

  return (st_mysqlnd_plugin_header*)instance.ptr;
}
