/*
  +----------------------------------------------------------------------+
  | Copyright (c) The PHP Group                                          |
  +----------------------------------------------------------------------+
  | This source file is subject to version 3.01 of the PHP license,      |
  | that is bundled with this package in the file LICENSE, and is        |
  | available through the world-wide-web at the following url:           |
  | http://www.php.net/license/3_01.txt                                  |
  | If you did not receive a copy of the PHP license and are unable to   |
  | obtain it through the world-wide-web, please send a note to          |
  | license@php.net so we can mail you a copy immediately.               |
  +----------------------------------------------------------------------+
  | Authors: Rafal Somla <rafal.somla@oracle.com>                        |
  +----------------------------------------------------------------------+
*/

#ifndef MYSQLND_KRB_GSSAPI_H
#define MYSQLND_KRB_GSSAPI_H

#include <gssapi/gssapi.h>
#include <gssapi/gssapi_krb5.h>

#include <string>
#include <string_view>
#include <optional>
#include <stdexcept>
#include <functional>
#include <cstddef>

namespace krb {
  // Note: used to print Kerberos specific error messages
  const char* get_error_msg(uint32_t);
}


namespace gssapi {

  using std::string;
  using std::string_view;
  using std::byte;
  using std::optional;


  /*
    Convenience wrapper around `gss_buffer_desc` structure that manages 
    proper releasing of resources.
  */

  struct Buffer
    : public gss_buffer_desc
  {
    /*
      Set to true for buffers allocated by GSSAPI library that need
      to be released.
    */

    bool free = false;

    Buffer(const gss_buffer_desc& other)
      : gss_buffer_desc{other}, free{true}
    {}

    Buffer(Buffer &&other)
      : gss_buffer_desc{other}
      , free{other.free}
    {
      other.free = false;
      other.value = nullptr;
    }

    Buffer(string_view data)
    {
      this->value = const_cast<void*>((const void*)data.data());
      this->length = data.length();
    }

    Buffer()
      // note: GSS_C_EMPTY_BUFFER macro contains braces as required
      : gss_buffer_desc GSS_C_EMPTY_BUFFER
    {}

    ~Buffer()
    {
      if (!free)
        return;

      // Note: we ignore errors here
      OM_uint32 minor;
      (void)gss_release_buffer(&minor, this);
      this->value = nullptr;
    }

    byte* data()
    {
      return (byte*)value;
    }
  };


  struct Context {

    /*
      Note: The 3rd argument is a function which is used to fetch credentials 
      (TGT) for the upn in case they can not be found in the cache.
    */

    Context(string_view spn, string upn
      , std::function<void(std::string)> get_cred = [](std::string){}
    );

    ~Context()
    {
      // Note: we ignore errors here
      OM_uint32 minor;
      (void)gss_release_name(&minor, &this->m_target);
      (void)gss_release_cred(&minor, &this->m_cred);
      (void)gss_delete_sec_context(&minor, &this->m_ctx, GSS_C_NO_BUFFER );
    }

    /*
      Generate next reply to a token from the peer in the handshake that 
      establishes the security context.

      Returns nothing if handshake has successfully ended and the context 
      is established. Otherwise returns buffer containing the reply that should 
      be sent to the peer and then another call to gen_token() should be made 
      when the peer sends its reply. Errors, if any, are thrown.
    */

    optional<Buffer> gen_token(Buffer in = {});

    bool m_started  = false;
    bool m_ready    = false;
    gss_ctx_id_t  m_ctx = GSS_C_NO_CONTEXT;
    gss_name_t    m_target;
    gss_OID       m_mech;
    gss_cred_id_t m_cred = GSS_C_NO_CREDENTIAL;
    OM_uint32     m_flags = 0;
  };


  // TODO: Cover Krb minor error codes, see KRB5_XXX constants in <krb5/krb5.h>

  inline
  const char* get_routine_error_msg(OM_uint32 major)
  {
    // see: https://datatracker.ietf.org/doc/html/rfc2744.html#section-5.21

    switch (GSS_ROUTINE_ERROR(major))
    {
    case GSS_S_BAD_MECH:
      return "An unsupported mechanism was requested";
    case GSS_S_BAD_NAME:
      return "An invalid name was supplied";
    case GSS_S_BAD_NAMETYPE:
      return "A supplied name was of an unsupported type";
    case GSS_S_BAD_BINDINGS:
      return "Incorrect channel bindings were supplied";
    case GSS_S_BAD_STATUS:
      return "An invalid status code was supplied";
    case GSS_S_BAD_MIC:
      return "A token had an invalid MIC";
    case GSS_S_NO_CRED:
      return "No credentials were supplied"
        ", or the credentials were unavailable or inaccessible.";
    case GSS_S_NO_CONTEXT:
      return "No context has been established";
    case GSS_S_DEFECTIVE_TOKEN:
      return "A token was invalid";
    case GSS_S_DEFECTIVE_CREDENTIAL:
      return "A credential was invalid";
    case GSS_S_CREDENTIALS_EXPIRED:
      return "The referenced credentials have expired";
    case GSS_S_CONTEXT_EXPIRED:
      return "The context has expired";
    case GSS_S_FAILURE:
      return "Miscellaneous failure";
    case GSS_S_BAD_QOP:
      return "The quality-of-protection requested could not be provided";
    case GSS_S_UNAUTHORIZED:
      return "The operation is forbidden by local security policy";
    case GSS_S_UNAVAILABLE:
      return "The operation or option is unavailable";
    case GSS_S_DUPLICATE_ELEMENT:
      return "The requested credential element already exists";
    case GSS_S_NAME_NOT_MN:
      return "The provided name was not a mechanism name";
    default:
      return nullptr;
    }
  }


  /*
    Analyse major and minor GSSAPI return codes and throw std::runtime_error{} 
    in case error is detected.
  */

  inline
  OM_uint32 check(OM_uint32 major, OM_uint32 minor, string_view ctx = {})
  {
    if (!GSS_ERROR(major))
      return major;

    DBG_FMT(
      "GSS ERROR: call: %d, routine: %d, suplementatry: %d, minor: %d",
      GSS_CALLING_ERROR_FIELD(major),
      GSS_ROUTINE_ERROR_FIELD(major),
      GSS_SUPPLEMENTARY_INFO(major),
      minor
    );

    // see: https://datatracker.ietf.org/doc/html/rfc2744.html#section-5.21

    try {
      switch (GSS_CALLING_ERROR(major))
      {
      case 0: break;
      case GSS_S_CALL_INACCESSIBLE_READ: throw "could not read input parameter";
      case GSS_S_CALL_INACCESSIBLE_WRITE: throw "could not write output parameter";
      case GSS_S_CALL_BAD_STRUCTURE: throw "malformed parameter";
      default: throw "unkonw calling error";
      }
    }
    catch (const char* msg)
    {
      DBG_FMT("Call error: %s", msg);
    }

    std::string msg{get_routine_error_msg(major) || "Unknown error"};

    if (GSS_S_FAILURE == GSS_ROUTINE_ERROR(major))
      msg = std::string{"Kerberos error: "} + krb::get_error_msg(minor);

    throw std::runtime_error{std::string{ctx} + ": " + msg};
  }


  inline
  Context::Context(
    string_view spn, string upn
    , std::function<void(std::string)> get_cred
  ) 
  {
    Buffer spn_buf{spn};
    Buffer upn_buf{upn};
    gss_name_t  uname;

    OM_uint32 minor;
    OM_uint32 major;

    // DBG_FMT("creating gssapi context");

    major = gss_import_name(
      &minor, &spn_buf, GSS_KRB5_NT_PRINCIPAL_NAME, &m_target
    );

    check(major, minor, "Error importing SPN");

    major = gss_import_name(&minor
      , &upn_buf, GSS_KRB5_NT_PRINCIPAL_NAME, &uname
    );

    check(major, minor, "Error importing UPN");

    /*
      Acquiring credentials.

      We make 2 attempts: first one should succeed if credentials for the upn 
      are already in the cache. If this is not the case we use the provided 
      get_cred() function to fetch credentials and make another attempt.
    */

    for (int attempt=0; attempt < 2; ++attempt)
    {
      DBG_FMT("acquire credentials, attempt %d", 1+attempt);

      major = gss_acquire_cred(&minor
        , uname
        , GSS_C_INDEFINITE // maximum permitted life-time
        , GSS_C_NO_OID_SET // TODO: select kerberos
        , GSS_C_INITIATE 
        , &m_cred
        , nullptr, nullptr
      );

      if (!GSS_ERROR(major))
      {
        DBG_FMT("got credentials for: %s", upn.c_str());
        return;
      }

      // If 2nd attempt failed we fail.

      if (attempt)
        throw std::runtime_error{"Could not fetch credentials for UPN"};

      switch (GSS_ROUTINE_ERROR(major))
      {
      case GSS_S_NO_CRED:
      case GSS_S_FAILURE: // internal Kerberos error
        DBG_FMT("Could not find credentials in cache, trying to fetch them");
        get_cred(upn);
        // FIXME: Possibility to not provide get_cred() and then no 2nd attempt.
        continue;

      default:
        // Any other error is a failure.
        check(major, minor, "Could not fetch credentials for UPN");
      }

      // Another attempt only if get_cred() was called above.
      break;
    }

  }


  inline
  optional<Buffer> Context::gen_token(Buffer input_token)
  {
    OM_uint32 minor;
    gss_buffer_desc output_token;

    DBG_FMT("gen_token(): %s, input size: %d",
      !m_started ? "init" : m_ready ? "completed" : "in progress",
      input_token.length
    );
    m_started = true;

    // see: https://datatracker.ietf.org/doc/html/rfc2744.html#section-5.19

    OM_uint32 ret = gss_init_sec_context(&minor
      , m_cred
      , &m_ctx
      , m_target
      , gss_mech_krb5  // explictly request Kerberos auth. mechanism
      , m_flags
      , 0 // time_req, 0 = default
      , GSS_C_NO_CHANNEL_BINDINGS
      , &input_token
      , &m_mech // actual_mech_type (gss_OID*) TODO - check that we use KRB
      , &output_token
      , nullptr // ret_flags (OM_uint32*)
      , nullptr // time_rec  (OM_uint32*) - number of sec. the context will be valid
    );

    ret = check(ret, minor, "gss_init_sec_context() error");

    if (ret & GSS_S_CONTINUE_NEEDED || 0 < output_token.length)
    {
      DBG_FMT("Continuing with output token of length %d", output_token.length);
      return output_token;
    }

    if (GSS_S_COMPLETE == ret)
    {
      DBG_FMT("Security context established using mechanism %d", m_mech);
      m_ready = true;
      return {};
    }

    // Note: This situation is forbidden by RFC#2743.
    throw std::logic_error{
      "gss_init_sec_context() returned invalid non-error code"
      " (neither complete nor continue)"
    };
  }


}  // gssapi namespace

#endif  // MYSQLND_KRB_GSSAPI_H